/**********************************************************************
 * Copyright (c) 2004 IBM Corporation and others. All rights reserved.   This
 * program and the accompanying materials are made available under the terms of
 * the Common Public License v1.0 which accompanies this distribution, and is
 * available at http://www.eclipse.org/legal/cpl-v10.html
 * 
 * Contributors: 
 * IBM - Initial API and implementation
 **********************************************************************/
package org.eclipse.core.resources.refresh.linux;

import java.util.HashMap;
import java.util.Map;

import org.eclipse.core.internal.refresh.RefreshManager;
import org.eclipse.core.resources.*;
import org.eclipse.core.resources.refresh.*;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.Path;
/**
 * Creates and returns linux file system monitors
 * 
 * @since 3.0
 */
public class LinuxRefreshProvider extends RefreshProvider {
	static final String DEBUG_PREFIX = "LinuxRefreshMonitor: "; //$NON-NLS-1$
	/**
	 * Map of the absolute path (String) of a resource to a HandlerContainer
	 */
	private final Map pathToContainerMap = new HashMap();

	private final LinuxReplyHandler replyJob = new LinuxReplyHandler(this);
	private final LinuxRequestHandler requestJob = new LinuxRequestHandler(this);
	private IRefreshResult result;

	void addHandle(String filePath, HandleContainer container) {
		boolean wasEmpty;
		synchronized (this) {
			wasEmpty= pathToContainerMap.isEmpty();
			pathToContainerMap.put(filePath, container);
		}
		System.out.println(DEBUG_PREFIX + "handle count: " + pathToContainerMap.size());
		//make sure the reply job is running
		if (wasEmpty)
			replyJob.schedule(LinuxReplyHandler.RESCHEDULE_DELAY);
	}
	/**
	 * FAM events can occur on both files and directories.  Since we don't
	 * install handlers on file, we need to obtain the monitor folder corresponding
	 * to the file.
	 * @param path a fully qualified resource path
	 * @param addFile
	 * @return The monitored resource, or <code>null</code>
	 */
	synchronized IResource getMonitoredResource(String path) {
		if (path == null || path.length() == 0)
			return null;
		HandleContainer container = (HandleContainer) pathToContainerMap.get(path);
		if (container != null)
			return container.resource;
		return null;
	}
	/* (non-Javadoc)
	 * @see org.eclipse.core.resources.refresh.RefreshProvider#installMonitor(org.eclipse.core.resources.IResource, org.eclipse.core.resources.refresh.IRefreshResult)
	 */
	public IRefreshMonitor installMonitor(IResource resource, IRefreshResult result) {
		this.result = result;
		if (resource.getLocation() == null || !resource.exists())
			return null;
		if (resource.getType() == IResource.FILE)
			return null;
		requestJob.addResource(resource);
		//make sure the reply job is running
		replyJob.schedule(LinuxReplyHandler.RESCHEDULE_DELAY);
		return requestJob;
	}
	/* (non-Javadoc)
	 * @see org.eclipse.core.resources.refresh.IRefreshResult#monitorFailed(org.eclipse.core.resources.refresh.IRefreshMonitor, org.eclipse.core.resources.IResource)
	 */
	public void monitorFailed(IResource resource) {
		if (result != null)
			result.monitorFailed(requestJob, resource);
	}
	/* (non-Javadoc)
	 * @see org.eclipse.core.resources.refresh.IRefreshResult#refresh(org.eclipse.core.resources.IResource)
	 */
	public void refresh(IResource resource) {
		if (result != null)
			result.refresh(resource);
	}
	LinuxRequestHandler getRequestHandler() {
		return requestJob;
	}
	synchronized HandleContainer getHandle(String filename) {
		return (HandleContainer)pathToContainerMap.get(filename);
	}
	boolean hasHandles() {
		return !pathToContainerMap.isEmpty();
	}
	synchronized void clearHandles() {
		pathToContainerMap.clear();
	}
	synchronized HandleContainer removeHandle(String path) {
		return (HandleContainer)pathToContainerMap.remove(path);
	}
}