/**********************************************************************
 * Copyright (c) 2004 IBM Corporation and others. All rights reserved.   This
 * program and the accompanying materials are made available under the terms of
 * the Common Public License v1.0 which accompanies this distribution, and is
 * available at http://www.eclipse.org/legal/cpl-v10.html
 * 
 * Contributors: 
 * IBM - Initial API and implementation
 **********************************************************************/
package org.eclipse.core.resources.refresh.linux;

import java.io.File;

import org.eclipse.core.internal.refresh.RefreshManager;
import org.eclipse.core.internal.utils.Policy;
import org.eclipse.core.resources.*;
import org.eclipse.core.runtime.*;
import org.eclipse.core.runtime.jobs.Job;

/**
 * Interaction with FAM happens in two separate jobs, since FAM will
 * block on requests if the reply queue becomes full.  This job handles
 * replies from FAM.
 * 
 * @since 3.0
 */
class LinuxReplyHandler extends Job {
	static final long RESCHEDULE_DELAY = 1000;
	private final LinuxRefreshProvider provider;

	LinuxReplyHandler(LinuxRefreshProvider provider) {
//		super(Policy.bind("LM.replyJobName")); //$NON-NLS-1$
		super("FAM Reply handler");
		this.provider = provider;
		setPriority(Job.DECORATE);
		setSystem(true);
		if (RefreshManager.DEBUG)
			System.out.println(LinuxRefreshProvider.DEBUG_PREFIX + " starting native monitor"); //$NON-NLS-1$
	}
	private IStatus processFAMEvents(IProgressMonitor monitor) {
		while (LinuxNatives.PendingEvents()) {
			if (monitor.isCanceled())
				return Status.CANCEL_STATUS;
			MonitorEvent event = LinuxNatives.GetNextEvent();
			if (event == null)
				continue;
			System.out.println(LinuxRefreshProvider.DEBUG_PREFIX + "received FAM event: " + event);
			switch (event.getEventType()) {
				case MonitorEvent.EVENT_TYPE_CHANGED:
					IResource toRefresh = provider.getMonitoredResource(event.getFilename());
					if (toRefresh != null)
						provider.refresh(toRefresh);
					break;
				case MonitorEvent.EVENT_TYPE_CREATED:
					File file = new File(event.getFilename());
					//ignore file creations as a change event will also occur on the folder
					if (!file.isDirectory())
						break;
					String parentPath = file.getParent();
					String name = file.getName();
					IResource parent = provider.getMonitoredResource(parentPath);
					if (parent != null) {
						provider.refresh(parent);
						IResource newFolder;
						// monitor the newly created path too, since monitors are not recursive
						if (parent.getType() == IResource.FOLDER)
							newFolder = ((IFolder)parent).getFolder(name);
						else //Project
							newFolder = ((IProject) parent).getFolder(name);
						provider.getRequestHandler().addResource(newFolder);
					}
					break;
				case MonitorEvent.EVENT_TYPE_DELETED:
					String filename = event.getFilename();
					toRefresh = provider.getMonitoredResource(filename);
					if (toRefresh != null) {
						provider.refresh(toRefresh);
						// stop monitoring this path, it doesn't exist anymore
						provider.getRequestHandler().unmonitor(toRefresh);
					}
					break;
				default:
					//the event is not known, so ignore it and check if there are any other events
			}
		}
		return Status.OK_STATUS;
	}
	/**
	 * We block on FAM events in here (not, note, polling FAM), so this runs in
	 * a separate thread
	 */
	protected IStatus run(IProgressMonitor monitor) {
//		if (RefreshManager.DEBUG)
//			System.out.println(LinuxRefreshProvider.DEBUG_PREFIX + " started " + getName()); //$NON-NLS-1$
		//always reschedule the job - so it will come back after errors or cancelation
		schedule(RESCHEDULE_DELAY);
		try {
			monitor.beginTask("", IProgressMonitor.UNKNOWN); //$NON-NLS-1$
			processFAMEvents(monitor);
		} finally {
			monitor.done();
//			if (RefreshManager.DEBUG)
//				System.out.println(LinuxRefreshProvider.DEBUG_PREFIX + " stopped " + getName()); //$NON-NLS-1$
		}
		return Status.OK_STATUS;
	}
	/* (non-Javadoc)
	 * @see org.eclipse.core.runtime.jobs.Job#shouldRun()
	 */
	public boolean shouldRun() {
		return provider.hasHandles();
	}
}