/**********************************************************************
 * Copyright (c) 2004 IBM Corporation and others. All rights reserved.   This
 * program and the accompanying materials are made available under the terms of
 * the Common Public License v1.0 which accompanies this distribution, and is
 * available at http://www.eclipse.org/legal/cpl-v10.html
 * 
 * Contributors: 
 * IBM - Initial API and implementation
 **********************************************************************/
package org.eclipse.core.resources.refresh.linux;

import java.util.ArrayList;
import java.util.List;

import org.eclipse.core.internal.refresh.RefreshManager;
import org.eclipse.core.internal.utils.Policy;
import org.eclipse.core.resources.*;
import org.eclipse.core.resources.refresh.IRefreshMonitor;
import org.eclipse.core.runtime.*;
import org.eclipse.core.runtime.jobs.Job;

/**
 * Interaction with FAM happens in two separate jobs, since FAM will
 * block on requests if the reply queue becomes full.  This job posts
 * requests to FAM.
 * 
 * @since 3.0
 */
class LinuxRequestHandler extends Job implements IRefreshMonitor {
	private static final long RESCHEDULE_DELAY = 3000;
	/**
	 * Any errors that have occurred
	 */
	private MultiStatus errors;
	private LinuxRefreshProvider provider;
	/**
	 * List of ResourceEvent that have not yet been processed
	 */
	private final List pendingEvents;
	/**
	 * If true, indicates a request to remove all active monitors
	 */
	private boolean unmonitorAll = false;

	LinuxRequestHandler(LinuxRefreshProvider provider) {
//		super(Policy.bind("LM.requestJobName")); //$NON-NLS-1$
		super("FAM request handler");
		this.provider = provider;
		setPriority(Job.DECORATE);
		setSystem(true);
		pendingEvents = new ArrayList();
		//add a startup event to be nice to other processes during startup
		pendingEvents.add(new ResourceEvent(null, ResourceEvent.STARTUP));
		if (RefreshManager.DEBUG)
			System.out.println(LinuxRefreshProvider.DEBUG_PREFIX + " starting native monitor"); //$NON-NLS-1$
	}
	private synchronized void addException(String message) {
		if (errors == null) {
			String msg = Policy.bind("LM.exceptions"); //$NON-NLS-1$
			errors = new MultiStatus(ResourcesPlugin.PI_RESOURCES, 1, msg, null);
		}
		errors.add(new Status(IStatus.ERROR, ResourcesPlugin.PI_RESOURCES, 1, message, null));
	}
	/**
	 * Adds the resource to the set of resources monitored by FAM.
	 * Resources are added to a collection of new resources, which are
	 * then processed asynchronously.
	 */
	public void addResource(IResource resource) {
		synchronized (pendingEvents) {
			pendingEvents.add(new ResourceEvent(resource, ResourceEvent.ADD));
		}
		//make sure the job is running 
		schedule(1000);
	}
	/**
	 *  Adds the resource to the set of resources monitored by FAM
	 */
	private void doAddResource(IResource resource) {
		IPath location = resource.getLocation();
		if (!resource.exists() || resource.getType() == IResource.FILE || location == null)
			return;
		String filePath = location.toOSString();
		nice();
		int handle = LinuxNatives.AddDirectoryMonitor(filePath);
		// Handle of 0 means the natives failed to add a monitor for this location
		if (handle == 0)  {
			provider.monitorFailed(resource);
			return;
		}
		// keep track of the handle so we can later remove it from monitoring
		HandleContainer hc = new HandleContainer(resource, handle);
		provider.addHandle(filePath, hc);
		if (RefreshManager.DEBUG)
			System.out.println(LinuxRefreshProvider.DEBUG_PREFIX + " added handle: " + hc ); //$NON-NLS-1$ //$NON-NLS-2$

		//recurse on child resources
		try {
			IResource[] children = ((IContainer) resource).members();
			for (int i = 0; i < children.length; i++) 
				doAddResource(children[i]);
		} catch (CoreException e) {
			//ignore inaccessible children
		}
	}
	/**
	 * It is important for this job not to hog CPU.  First, because it interferes
	 * with activity in the main thread, and second, because FAM cannot
	 * handle a large number of consecutive requests due to buffer
	 * overrun problems.
	 */
	private void nice() {
		try {
			Thread.sleep(25);
			Thread.yield();
		} catch (InterruptedException e) {
			//allow interruption
		}
	}
	/**
	 * Stop monitoring path
	 * 
	 * @param path path to be removed from monitoring
	 */
	private void doRemoveResource(IResource resource) {
		IPath location = resource.getLocation();
		if (location == null)
			return;
		if (RefreshManager.DEBUG)
			System.out.println(LinuxRefreshProvider.DEBUG_PREFIX + "closing monitor for: " + resource); //$NON-NLS-1$
		HandleContainer container = provider.removeHandle(location.toOSString());
		if (container == null) {
			if (RefreshManager.DEBUG)
				System.out.println(LinuxRefreshProvider.DEBUG_PREFIX + "missing handle for path: " + location.toOSString()); //$NON-NLS-1$
			return;
		}
		nice();
		if (!LinuxNatives.RemoveMonitor(container.handle))
			addException(Policy.bind("LM.err_close", resource.toString())); //$NON-NLS-1$
		if (RefreshManager.DEBUG)
			System.out.println(LinuxRefreshProvider.DEBUG_PREFIX + "removed handle: " + container); //$NON-NLS-1$
	}
	/**
	 * Remove all FAM monitors and close connection
	 */
	private void doRemoveAll() {
		//clear the container set and cancel the job
		provider.clearHandles();
		//remove native monitors
		if (!LinuxNatives.FAMClose())
			addException(Policy.bind("PM.err_close", "FAMClose")); //$NON-NLS-1$ //$NON-NLS-2$
	}
	/**
	 * Asynchronously process pending groups of resources that need
	 * to be added to FAM or removed from FAM.
	 */
	private void processEvents() {
		ResourceEvent[] events = null;
		synchronized (pendingEvents) {
			if (!pendingEvents.isEmpty())
				events = (ResourceEvent[]) pendingEvents.toArray(new ResourceEvent[0]);
			pendingEvents.clear();
		}
		if (events != null) {
			for (int i = 0; i < events.length; i++) {
				System.out.println(LinuxRefreshProvider.DEBUG_PREFIX + "received resource event: " + events[i]);
				switch (events[i].kind) {
					case ResourceEvent.ADD:
						doAddResource(events[i].resource);
						break;
					case ResourceEvent.REMOVE:
						doRemoveResource(events[i].resource);
						break;
					case ResourceEvent.REMOVE_ALL:
						doRemoveAll();
						break;
					case ResourceEvent.STARTUP:
						try {
							Thread.sleep(20000);
						} catch (InterruptedException e) {
							// allow interruption
						}
						break;
				}
			}
		}
	}
	/* (non-Javadoc)
	 * @see org.eclipse.core.internal.jobs.InternalJob#run(org.eclipse.core.runtime.IProgressMonitor)
	 */
	protected IStatus run(IProgressMonitor monitor) {
		if (RefreshManager.DEBUG)
			System.out.println(LinuxRefreshProvider.DEBUG_PREFIX + " started " + getName()); //$NON-NLS-1$
		//always reschedule the job - so it will come back after errors or cancelation
		schedule(RESCHEDULE_DELAY);
		try {
			monitor.beginTask("", IProgressMonitor.UNKNOWN); //$NON-NLS-1$
			processEvents();
		} finally {
			monitor.done();
			if (RefreshManager.DEBUG)
				System.out.println(LinuxRefreshProvider.DEBUG_PREFIX + " stopped " + getName()); //$NON-NLS-1$
		}
		MultiStatus result = errors;
		errors = null;
		if (result != null && !result.isOK())
			return result;
		return Status.OK_STATUS;
	}
	/* (non-Javadoc)
	 * @see org.eclipse.core.runtime.jobs.Job#shouldRun()
	 */
	public boolean shouldRun() {
		return !pendingEvents.isEmpty();
	}
	/* (non-Javadoc)
	 * @see org.eclipse.core.resources.refresh.IRefreshMonitor#unmonitor(org.eclipse.core.resources.IResource)
	 */
	public void unmonitor(IResource resource) {
		//let monitor removals be processed asynchronously
		ResourceEvent event;
		if (resource != null)
			event = new ResourceEvent(resource, ResourceEvent.REMOVE);
		 else
			event = new ResourceEvent(null, ResourceEvent.REMOVE_ALL);
		synchronized (pendingEvents) {
			pendingEvents.add(event);
		}
		//make sure the job is running 
		schedule(RESCHEDULE_DELAY);
	}
}