/*******************************************************************************
 * Copyright (c) 2000, 2003 IBM Corporation and others.
 * All rights reserved. This program and the accompanying materials 
 * are made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 * 
 * Contributors:
 *     IBM Corporation - initial API and implementation
 *******************************************************************************/
package org.eclipse.help.internal.util;

import java.io.*;
import java.net.*;
import java.util.*;

import org.eclipse.core.runtime.*;
import org.eclipse.help.*;
import org.eclipse.help.internal.*;
import org.osgi.framework.Bundle;

public class ResourceLocator {
	private static final Hashtable zipCache = new Hashtable();
	private static final Object ZIP_NOT_FOUND = new Object();
	// Indicates there is no dynamic content provider for a particular plugin
	private static final Object STATIC_DOCS_ONLY = ZIP_NOT_FOUND;
	// Map of document content providers by plug-in ID;
	private static Map contentProducers = new HashMap(2, 0.5f);

	/**
	 * Obtains content proivider for a documentation plug-in,
	 * creates one if necessary.
	 * @param pluginId
	 * @return ITopicContentProvider or null
	 */
	private static IHelpContentProducer getContentProducer(Bundle plugin) {
		Object producer = contentProducers.get(plugin);
		if (producer == null) {
			// first time for the plug-in, so attempt to
			// find and instantiate provider
			producer = createContentProducer(plugin);
			if (producer == null) {
				producer = STATIC_DOCS_ONLY;
			}
			contentProducers.put(plugin, producer);
		}
		if (producer == STATIC_DOCS_ONLY) {
			return null;
		} else {
			return (IHelpContentProducer) producer;
		}

	}

	/**
	 * Creates content proivider for a documentation plug-in
	 * @param pluginId
	 * @return ITopicContentProvider or null
	 */
	private static IHelpContentProducer createContentProducer(Bundle plugin) {
		IExtensionPoint xp =
			Platform.getExtensionRegistry().getExtensionPoint(
				"org.eclipse.help.contentProducer");
		if (xp == null) {
			return null;
		}
		IExtension[] extensions = xp.getExtensions();
		for (int i = 0; i < extensions.length; i++) {
			if (!extensions[i].getNamespace().equals(plugin.getSymbolicName())) {
				continue;
			}
			IConfigurationElement[] elements =
				extensions[i].getConfigurationElements();
			for (int j = 0; j < elements.length; j++) {
				if (!"contentProducer".equals(elements[j].getName())) {
					continue;
				}
				try {
					Object o = elements[j].createExecutableExtension("producer");
					if (o instanceof IHelpContentProducer) {
						return (IHelpContentProducer) o;
					}
				} catch (CoreException ce) {
					HelpPlugin.logError(
						HelpResources.getString(
							"E044",
							plugin.getSymbolicName()),
						ce);
				}
			}
		}
		return null;
	}

	/**
	 * Opens an input stream to a file contained in a plugin.
	 * This includes NL lookup.
	 */
	public static InputStream openFromProducer(
		Bundle pluginDesc,
		String file,
		String locale) {
		IHelpContentProducer producer = getContentProducer(pluginDesc);
		if (producer == null) {
			return null;
		}
		if (locale == null || locale.length() <= 0) {
			locale = Platform.getNL();
		}
		Locale l;
		if (locale.length() >= 5) {
			l = new Locale(locale.substring(0, 2), locale.substring(3, 5));
		} else if (locale.length() >= 2) {
			l = new Locale(locale.substring(0, 2), "");
		} else {
			l = Locale.getDefault();
		}
		return producer.getInputStream(pluginDesc.getSymbolicName(), file, l);
	}

	/**
	 * Opens an input stream to a file contained in a zip in a plugin.
	 * This includes NL lookup.
	 */
	public static InputStream openFromZip(
		String pluginId,
		String zip,
		String file,
		String locale) {
		return openFromZip(Platform.getBundle(pluginId), zip, file, locale);
	}

	/**
	 * Opens an input stream to a file contained in a plugin.
	 * This includes NL lookup.
	 */
	public static InputStream openFromPlugin(
		String pluginId,
		String file,
		String locale) {
		return openFromPlugin(Platform.getBundle(pluginId), file, locale);
	}

	/**
	 * Opens an input stream to a file contained in a zip in a plugin.
	 * This includes NL lookup.
	 */
	public static InputStream openFromZip(
		Bundle pluginDesc,
		String zip,
		String file,
		String locale) {
		// First try the NL lookup
		InputStream is = doOpenFromZip(pluginDesc, "$nl$/" + zip, file, locale);
		if (is == null)
			// Default location <plugin>/doc.zip
			is = doOpenFromZip(pluginDesc, zip, file, locale);
		return is;
	}

	/**
	 * Opens an input stream to a file contained in a plugin.
	 * This includes NL lookup.
	 */
	public static InputStream openFromPlugin(
		Bundle pluginDesc,
		String file,
		String locale) {
		InputStream is = doOpenFromPlugin(pluginDesc, "$nl$/" + file, locale);
		if (is == null)
			// Default location
			is = doOpenFromPlugin(pluginDesc, file, locale);
		return is;
	}

	/**
	 * Opens an input stream to a file contained in doc.zip in a plugin
	 */
	private static InputStream doOpenFromZip(
		Bundle pluginDesc,
		String zip,
		String file,
		String locale) {
		String realZipURL = findZip(pluginDesc, zip, locale);
		if (realZipURL == null) {
			return null;
		}
		try {
			URL jurl = new URL("jar", "", realZipURL + "!/" + file);

			URLConnection jconnection = jurl.openConnection();
			jconnection.setDefaultUseCaches(false);
			jconnection.setUseCaches(false);
			return jconnection.getInputStream();

		} catch (IOException ioe) {
			return null;
		}
	}

	/**
	 * Opens an input stream to a file contained in a plugin
	 */
	private static InputStream doOpenFromPlugin(
		Bundle pluginDesc,
		String file,
		String locale) {
		IPath flatFilePath = new Path(file);
		Map override = new HashMap(1);
		override.put("$nl$", locale);
		URL flatFileURL = Platform.find(pluginDesc, flatFilePath, override);
		if (flatFileURL != null)
			try {
				return flatFileURL.openStream();
			} catch (IOException e) {
				return null;
			}
		return null;
	}
	/**
	 * @param pluginDesc
	 * @param zip zip file path as required by Plugin.find()
	 * @param locale
	 * @return String form of resolved URL of a zip or null
	 */
	private static String findZip(
		Bundle pluginDesc,
		String zip,
		String locale) {
		String pluginID = pluginDesc.getSymbolicName();
		// check cache
		Object cached = zipCache.get(pluginID + '/' + zip + '/' + locale);
		if (cached == null) {
			// not in cache find on filesystem
			IPath zipFilePath = new Path(zip);
			Map override = new HashMap(1);
			override.put("$nl$", locale);
			try {
				URL zipFileURL =
					Platform.find(pluginDesc, zipFilePath, override);	//PASCAL This will not activate the plugin
				if (zipFileURL != null) {
					URL realZipURL = Platform.asLocalURL(Platform.resolve(zipFileURL));
					cached = realZipURL.toExternalForm();
				} else {
					cached = ZIP_NOT_FOUND;
				}
			} catch (IOException ioe) {
				cached = ZIP_NOT_FOUND;
			}
			// cache it
			zipCache.put(pluginID + '/' + zip + '/' + locale, cached);
		}
		if (cached == ZIP_NOT_FOUND) {
			return null;
		}
		return (String) cached;
	}

}
